/*
 * Project: PAPI: Perfometer
 * File:    perfometer.c
 * Author:  Phillip Mucci
 *          mucci@cs.utk.edu
 * Date:    ???
 *
 * Mods:    Nathan P. Garner
 *          garner@cs.utk.edu
 * Date:    Jan. 2000
 *
 *          Kevin London
 *          london@cs.utk.edu
 *          July 2000
 *          Changed communication for the new server that allows
 *          multiple processes to be monitored.
 *
 *          Kevin London
 *          london@cs.utk.edu
 *          October 2000
 *          Added MPI support
 *
 *          Kevin London
 *          london@cs.utk.edu
 *          Jan. 2001
 *          Changed the protocol to be asynch, optimized the buffers
 *          added support for alerts, pause of execution and more.
 *
 *          Kevin London
 *          london@cs.utk.edu
 *          Feb. 2001
 *          Change of metric support added.
 *
 *	   Kevin London
 *	   london@cs.utk.edu
 *	   Feb 2001
 *	   Code has gotten a little buggy and too many band-aids
 *	   this is a rewrite of the code to be restructured,
 *	   more stable and easier to maintain.
 *
 * 	   Kevin London
 *	   london@cs.utk.edu
 *	   November 2001
 *	   Added Windows 2000 support, this is non-blocking
 *	   but not asynch as windows doesn't support asynch
 *	   communication (Or at least I can't find it ;) )
 *
 *	   Kevin London
 *	   london@cs.utk.edu
 *	   November 2001
 *	   Added a NO_ASYNCH so if your platform has a problem
 * 	   with asynchronus sockets, you can use non-blocking
 *	   sockets.
 *
 *	   Philip Mucci
 *	   mucci@cs.utk.edu
 *	   August 2002
 *         Added thread safe interface for use with Dynaprof
 */

#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <memory.h>
#include <assert.h>
#include <sys/types.h>
#include <limits.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#ifndef WIN32
#include <unistd.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <pwd.h>
#include <netdb.h>
#include <stropts.h>
#else 
#include <windows.h>
#include <winbase.h>
#endif
#ifdef SOLARIS
#include <sys/file.h>
#endif
#ifdef USE_MPI
#include <mpi.h>
#endif
#include "papi.h"
#include "perfometer.h"
#include "perfometer_server.h"
#include "conn.h"

#define socklen_t int

/*
 * Definitions
 */
#define FALSE 0
#define TRUE 1

#define PERF_PORTNUM 3545  /* Randomly selected port number */

#define TIMER_RESTART     0
#define TIMER_PAUSE       1
#define TIMER_NEWVALUE    2
#define TIMER_DEFAULT     3

#define PAPI_PERFOMETER_OPTS "PAPI_PERFOMETER_OPTS"
#define PAPI_PERFOMETER_NAME "Perfometer"
#define PAPI_PERFOMETER_PATH "PAPI_PERFOMETER_PATH"
#if (REAL_TIMER|USE_MPI)
#define PERFOMETER_MS 150
#elif WIN32
#define PERFOMETER_MS 200
#else
#define PERFOMETER_MS 50
#endif
#define PERFOMETER_SEC 0

#ifdef WIN32 
	// Win2000: Indicates a process that has priority above
	// IDLE_PRIORITY_CLASS but below NORMAL_PRIORITY_CLASS.
	#define BELOW_NORMAL_PRIORITY_CLASS		0x4000
	// Win2000: Indicates a process that has priority above
	// NORMAL_PRIORITY_CLASS but below HIGH_PRIORITY_CLASS.
	#define ABOVE_NORMAL_PRIORITY_CLASS		0x8000
#endif


/*
 * Global Declerations
 */

struct perfometer_data 
{
  const PAPI_preset_info_t *info;
  const PAPI_exe_info_t *exeinfo;
  const PAPI_hw_info_t *hwinfo;
  char buf[250];
  float real_time;
  long_long start_us, vstart_us;
#ifndef WIN32
  uid_t my_uid;
  char my_hostname[MAXHOSTNAMELEN+1];
  int readsockfd, sockfd, read_fd, write_fd;
  struct sigaction action, read_action;
  struct itimerval value;
  long long values[8];
#else
  int my_uid;
  char my_hostname[256];
  SOCKET readsockfd, sockfd, read_fd, write_fd;
  UINT value;
  LONGLONG values[8];
#endif
  char *user_name;
  char *my_hostip,label[256];
  int EventSet, data_stream;
  int paused_app, num_data_values,color, events[8];
  float mhz;
#ifdef USE_MPI
  int rank, size,usempi;
#endif
}; 

/*
 *  Function declerations
 */
int perfometer_internal();
static int change_timer( struct perfometer_data *, int flag, int usec );
static char * strip_word( char *buf, char *buf_first, char delim );
static int read_mesg( struct perfometer_data *, int filedes, char *mesg );
static int write_mesg( struct perfometer_data *perf, int filedes, int type, char *mesg );
static void perfometer_check_messages_async(int sig, siginfo_t *, void *);
#if (WIN32|NO_ASYNCH)
static void perfometer_check_messages_noasync(struct perfometer_data *perf);
#endif
#ifndef WIN32
static void perfometer_handler( int sig );
#else
STATIC VOID CALLBACK perfometer_handler ( UINT wTimerID, UINT msg, DWORD dwuser, DWORD dw1, DWORD dw2 );
#endif
static int perfometer_change_event( struct perfometer_data *perf, char *mesg );
#ifdef SPEEDSTEP
static float recalc_mhz();
#endif

#ifndef WIN32
#if (REAL_TIMER|USE_MPI)
const static int   perfometer_timer  = ITIMER_REAL;
const static int   perfometer_signal = SIGALRM;
#else
const static int   perfometer_timer  = ITIMER_VIRTUAL;
const static int   perfometer_signal = SIGVTALRM;
#endif
#ifndef NO_ASYNCH
const static int   perfometer_readsig = SIGIO;
#endif
#endif

#ifndef THREADS
static struct perfometer_data *default_perfometer_data = NULL;
#endif

/*
 * internal_perfometer - This function initializes all necessary
 * information for perfometer and does a handshake with a process
 * on the socket to verify and transfer necessary initialization
 * information.  This function also setsup and installs a
 * signal handler, and virtual timer.  All communication in this
 * function is BLOCKING, after this function communication is
 * NON-BLOCKING. -KSL
 */

void * _internal_perfometer(int color, char *color_name ) {
   int actualport=0,readport=0,i;
   struct perfometer_data perf;
#ifndef WIN32
   size_t sin_size;
   struct passwd *pwtmp;
#else
   int sin_size;
    WORD wVersionRequested;
    WSADATA wsaData;
	HANDLE myProcessHandle;
#endif
   struct sockaddr_in their_addr;
   time_t when;
   struct hostent *myhost;
   struct in_addr in;
   char tmpbuf[500];
   char buf[3000];
#ifdef USE_MPI
   MPI_Status status;
   int port[12],tmp;
   char *name[12];
#endif

#ifdef WIN32
    wVersionRequested = MAKEWORD( 2, 0 );
    WSAStartup ( wVersionRequested, &wsaData );
#endif

   sin_size = sizeof(struct sockaddr_in);
   /* Setup defaults */
   perf.EventSet = PAPI_NULL;
   perf.color = 0;
   perf.real_time=0;
   perf.hwinfo = NULL;
   perf.exeinfo=NULL;
   perf.info=NULL;
   perf.paused_app = 0;
   perf.data_stream = 1;
#ifdef USE_MPI
   perf.usempi=0;
#endif

   sprintf( perf.label, "%s", "Unknown");
   if ( PAPI_library_init( PAPI_VER_CURRENT ) != PAPI_VER_CURRENT ) {
	fprintf( stderr, "Unable to initialize PAPI library, continuing without PAPI instrumentation.\n");
	return NULL;
   }
   if ( ( perf.hwinfo = PAPI_get_hardware_info() ) == NULL ) {
	fprintf( stderr, "Unable to get hardware information, continuing without PAPI instrumentation.\n");
	return NULL;
   }
   if ( ( perf.exeinfo = PAPI_get_executable_info() ) == NULL ) {
	fprintf( stderr, "Unable to get executable information, continuing without PAPI instrumentation.\n");
	return NULL;
   }   
   if ( ( perf.info = PAPI_query_all_events_verbose() ) == NULL ) {
	fprintf( stderr, "Unable to get event information, continuing without PAPI instrumentation.\n");
	return NULL;
   }
   when = time(NULL);
   if ( when == (time_t)-1 ) return NULL;
#ifndef WIN32
   perf.my_uid = getuid();
   pwtmp = getpwuid( perf.my_uid );
   if ( !pwtmp ) {
        perf.user_name = malloc( sizeof(char)*8);
        sprintf( perf.user_name, "%d", PERF_ANY );
   }
   else
        perf.user_name = (char *) strdup(pwtmp->pw_name);
#else
   /* drop the priority of this process to give the GUI more bandwidth */
   myProcessHandle = GetCurrentProcess();
   SetPriorityClass(myProcessHandle, BELOW_NORMAL_PRIORITY_CLASS);

   perf.user_name = malloc( sizeof(char)*8);
   sprintf( perf.user_name, "%d", PERF_ANY );
   perf.my_uid = 1;
#endif


#ifndef WIN32
   if ( gethostname( perf.my_hostname, MAXHOSTNAMELEN ) == -1 )
#else
   if ( gethostname( perf.my_hostname, 255) == -1 )
#endif
	sprintf( perf.my_hostname, "<unknown>");
   if ( (myhost=gethostbyname(perf.my_hostname)) ) {
	memcpy( &in.s_addr, *myhost->h_addr_list, sin_size );
	perf.my_hostip = (char *) strdup ( inet_ntoa(in) );
   }
   else
	perf.my_hostip = (char *) strdup("0.0.0.0");

   /* Setup port connections */
   if ( (actualport = setportconn( &(perf.sockfd), PERF_PORTNUM, 0 )) != PERF_PORTNUM)
   {
        if ( actualport < PERF_PORTNUM || actualport > (PERF_PORTNUM+5) )
                actualport = setportconn( &(perf.sockfd), 0, 0 );
   }


   if ((readport=setportconn(&(perf.readsockfd),PERF_PORTNUM+1,0))!=PERF_PORTNUM+1)
        readport = setportconn( &(perf.readsockfd), 0, 0);

#ifndef USE_MPI
    fprintf(stdout,"Perfometer client awaiting connection on port #%d.\n", actualport );
    fflush(stdout);
#else
   MPI_Initialized( &(perf.usempi) );
   if ( perf.usempi ) {
#ifdef DEBUG
        fprintf( stdout, "MPI is initialized, using MPI protocol.\n");
#endif
        MPI_Comm_rank( MPI_COMM_WORLD, &(perf.rank) );
        MPI_Comm_size( MPI_COMM_WORLD, &(perf.size) );
        if ( perf.size > 12 ) {
                fprintf(stdout,"Perfometer does not currently support a MPI job larger than 12 proceesses.\n" );
                fprintf(stdout, "Continuing application without PAPI instrumentation.\n");
                return NULL;
        }
        if ( perf.rank == 0 ) {
                fprintf(stdout, "MPI Perfometer client awaiting connection on port#%d.\n", actualport );
                fflush(stdout);
                port[0] = actualport;
                name[0] = strdup( perf.my_hostname );
                for ( i = 1; i<perf.size;i++ ) {
                    MPI_Recv( buf, MAX_BUFFER_SIZE, MPI_CHAR, i, 3545,
                        MPI_COMM_WORLD, &status );
                    MPI_Recv( &tmp, 1, MPI_INT, i, 3546, MPI_COMM_WORLD,
                        &status );
                    port[i] = tmp;
                    name[i] = strdup( buf );
#ifdef DEBUG
                    fprintf(stdout, "Master Recieved: %d %s\n", tmp, buf );
#endif
                }
        }
        else {
                MPI_Send( perf.my_hostname, strlen(perf.my_hostname),MPI_CHAR,
			0, 3545, MPI_COMM_WORLD );
                MPI_Send( &actualport, 1, MPI_INT, 0, 3546, MPI_COMM_WORLD );
#ifdef DEBUG
                fprintf(stdout, "Slave %d sending %s %d\n", perf.rank,
                        perf.my_hostname, actualport );
#endif
        }
    }
    else
        fprintf(stdout,"Perfometer client awaiting connection on port #%d.\n",actualport );
        fflush(stdout);
#endif
   /* Handshake */
   for(;;) {
        if ( perf.write_fd ){
#ifndef WIN32
                close ( perf.write_fd );
#else
                closesocket( perf.write_fd );
#endif
                perf.write_fd = 0;
        }
        if (perf.read_fd ) {
#ifndef WIN32
                close( perf.read_fd );
#else
                closesocket( perf.read_fd );
#endif
                perf.read_fd = 0;
        }
#ifdef DEBUG
        fprintf( stdout, "Awaiting connection on port: %d\n", actualport );
#endif
        if ( ( perf.write_fd = accept( perf.sockfd, (struct sockaddr*)&their_addr, &sin_size)) == -1 ) {
                fprintf( stdout, "Error setting up socket, continuing application without PAPI instrumentation.\n");
                return NULL;
        }
#ifdef DEBUG
        fprintf( stdout, "Got connection!\n" );
#endif

    /*
     *  Lets make sure we are talking to a PAPI connection and not a port
     *  scan or something else.
     */
    sprintf( buf, "%s%s%s%s%s%d%s%d" , PERF_DELIM,
           perf.user_name, PERF_DELIM, perf.my_hostip,
           PERF_DELIM, PERF_NONEEDAPP, PERF_DELIM, 1);

    if ( (write_mesg(&perf, perf.write_fd, PERF_APP, buf )) < 0 ) continue;
#ifdef DEBUG
    fprintf(stdout,"Perfometer server: got connection from %s\n", inet_ntoa(their_addr.sin_addr));
#endif
    if (  read_mesg(&perf,perf.write_fd, buf ) != PERF_ACK ) continue;

    if ( ( write_mesg(&perf, perf.write_fd, readport, NULL)) < 0 ) continue;
    if ((perf.read_fd = accept(perf.readsockfd, (struct sockaddr *)&their_addr, &sin_size)) == -1) {
         fprintf( stderr, "Error accepting socket\n");
         continue;
    }


    /* Setup new buffer sizes */
#ifndef WIN32
    tcp_set_option( perf.sockfd, SOL_SOCKET, SO_SNDBUF, 524288, "Sendbuffer");
    tcp_set_option( perf.readsockfd, SOL_SOCKET, SO_RCVBUF, 128, "Recvbuffer");
#endif

    if ( read_mesg(&perf, perf.read_fd, buf ) != PERF_ACK ) continue;

#ifdef USE_MPI
       if ( perf.usempi ) {
         buf[0] = '\0';
         if ( perf.rank == 0 ) {
           for ( i = 1; i < perf.size; i++ ) {
             sprintf(tmpbuf, "%s%s%s%d", PERF_DELIM, name[i], PERF_DELIM, port[i]);
             strcat( buf, tmpbuf );
             free(name[i]);
           }
           if ( ( write_mesg(&perf, perf.write_fd, PERF_ACK, buf ) ) < 0 ) continue;
         }
         else
           if ( (write_mesg(&perf, perf.write_fd, PERF_ACK, NULL )) < 0 ) continue;
         }
       else
          if ( (write_mesg(&perf, perf.write_fd, PERF_ACK, NULL )) < 0 ) continue;
#else
       if ( (write_mesg(&perf, perf.write_fd, PERF_ACK, NULL )) < 0 ) continue;
#endif
    /* Send the hardware info */
    sprintf(buf, "%s%d%s%s%s%d%s%s%s%d%s%d%s%d%s%d%s%s%s%d%s%s%s%f%s%f",
        PERF_DELIM,
        perf.my_uid, PERF_DELIM, "0",PERF_DELIM,PERF_HWINFO, PERF_DELIM,
        perf.my_hostname, PERF_DELIM, perf.hwinfo->ncpu, PERF_DELIM,
        perf.hwinfo->nnodes, PERF_DELIM, perf.hwinfo->totalcpus, PERF_DELIM,
        perf.hwinfo->vendor, PERF_DELIM, perf.hwinfo->vendor_string,
        PERF_DELIM, perf.hwinfo->model, PERF_DELIM, perf.hwinfo->model_string,
        PERF_DELIM, perf.hwinfo->revision, PERF_DELIM, perf.hwinfo->mhz );

    perf.mhz = perf.hwinfo->mhz;
    if ( (write_mesg(&perf,perf.write_fd,PERF_DATA,buf)) < 0 ) continue;

    /* Send the executable info but don't send the start/end addresses
       of the text, data and bss segments */
    sprintf( buf, "%s%d%s%s%s%d%s%s%s%s%s%s",
        PERF_DELIM, perf.my_uid, PERF_DELIM, "0",
        PERF_DELIM, PERF_EXEINFO, PERF_DELIM, perf.exeinfo->fullname,
        PERF_DELIM, perf.exeinfo->name, PERF_DELIM,
        (!perf.exeinfo->lib_preload_env
	||perf.exeinfo->lib_preload_env[0]=='\0')?
	"Default":perf.exeinfo->lib_preload_env);
    if ( (write_mesg(&perf,perf.write_fd,PERF_DATA,buf)) < 0 ) continue;

    sprintf( buf, "%s%d%s%s%s%d", PERF_DELIM, perf.my_uid, PERF_DELIM, "0",
        PERF_DELIM, PERF_EVENTLIST);
    for ( i=0; i<PAPI_MAX_PRESET_EVENTS;i++ ) {
        if ( perf.info[i].event_name && perf.info[i].avail ) {
          sprintf( tmpbuf, "%s%s%s%s", PERF_DELIM, perf.info[i].event_name,
                PERF_DELIM, perf.info[i].event_descr );
          strcat( buf, tmpbuf );
        }
    }
    if ( (write_mesg(&perf,perf.write_fd, PERF_DATA, buf )) < 0 ) continue;
    if (  read_mesg(&perf,perf.read_fd, buf ) != PERF_ACK ) continue;
    break;
   }


   /*
    * Wait here for the eventset we are suppose to monitor
    */
   while ( read_mesg(&perf, perf.read_fd, buf ) != PERF_CHANGE_EVENT );
   perf.start_us =  PAPI_get_real_usec();
   perf.vstart_us = PAPI_get_virt_usec();

#ifdef USE_MPI /* Wait for everyone to catch up */
   MPI_Barrier(MPI_COMM_WORLD);
#endif
#ifdef DEBUG
   fprintf(stdout, "Event changed, continuing.....\n");
#endif
#ifndef WIN32  /* Ok we can't do asynch for perfometer yet...... -KSL */
#ifndef NO_ASYNCH
   /*
    * Ok turn on the nonblocking bits, setup timer and the signal handlers
    * except for Non-asynchronus communication
    */
   setnonblocking( perf.sockfd );
   setnonblocking( perf.readsockfd );
   if ( fcntl( perf.read_fd, F_SETOWN, getpid())<0){
        perror("fcntl()");
        fprintf(stderr,"Continuing without PAPI instrumentation.\n");
        return NULL;
   }
   if ( fcntl( perf.read_fd, F_SETFL, FASYNC) < 0 ) {
        perror("fcntl()");
        fprintf(stderr,"Continuing without PAPI instrumentation.\n");
        return NULL;
   }

   if ( signal(perfometer_readsig, SIG_IGN) != SIG_DFL ) {
        fprintf(stderr,"User signal handler detected, perfometer disabled.\n");
        PAPI_stop( perf.EventSet, perf.values );
        return NULL;
   }
   memset ( &(perf.read_action), 0x00, sizeof(struct sigaction));
   perf.read_action.sa_sigaction = perfometer_check_messages_async;
   perf.read_action.sa_flags = SA_RESTART | SA_SIGINFO;
   if ( sigaction( perfometer_readsig, &(perf.read_action), NULL ) < 0 ) {
        perror("sigaction()");
        PAPI_stop( perf.EventSet, perf.values );
        return NULL;
   }
#endif /* NO_ASYNCH */
   if ( signal(perfometer_signal, SIG_IGN) != SIG_DFL){
        fprintf(stderr,"User signal handler detected, perfometer disabled.\n");
        PAPI_stop( perf.EventSet, perf.values );
        return NULL;
   }
   memset( &(perf.action), 0x00,sizeof(struct sigaction ) );
   perf.action.sa_handler = perfometer_handler;
   perf.action.sa_flags = SA_RESTART;
   if ( sigaction( perfometer_signal, &(perf.action), NULL ) < 0 ) {
        printf("Perfometer sigaction error!!!\n");
        perror("sigaction()");
        PAPI_stop( perf.EventSet, perf.values );
        return NULL;
   }
#endif
   perf.paused_app = 0;
   change_timer (&perf, TIMER_DEFAULT, 0 );

#ifdef DEBUG
   fprintf(stdout, "Past initialization and handshaking!\n");
#endif

  {
    struct perfometer_data *n = malloc(sizeof(perf));
    if (n == NULL)
      return NULL;
    memcpy(n,&perf,sizeof(perf));
    _internal_mark_perfometer(n,color,color_name);
#ifdef THREADS
    store_thread_specific_data(n);
#else
    default_perfometer_data = n;
#endif
    return n;
  }
}


/*
 * Reset the timer
 */
int change_timer( struct perfometer_data *perf, int flag, int newsec ) {
#ifndef WIN32 /* Ok we need a timer for windows */
  static int tv_sec=0, tv_usec=0,vtv_sec=0,vtv_usec=0;
  switch ( flag ) {
     case TIMER_PAUSE:
        tv_sec = perf->value.it_interval.tv_sec;
        tv_usec = perf->value.it_interval.tv_usec;
        vtv_sec = perf->value.it_value.tv_sec;
        vtv_usec = perf->value.it_value.tv_usec;
        perf->value.it_interval.tv_sec = 0;
        perf->value.it_interval.tv_usec = 0;
        perf->value.it_value.tv_sec = 0;
        perf->value.it_value.tv_usec=0;
        break;
    case TIMER_DEFAULT:
        perf->value.it_interval.tv_sec = PERFOMETER_SEC;
        perf->value.it_interval.tv_usec =PERFOMETER_MS*1000;
        perf->value.it_value.tv_sec = PERFOMETER_SEC;
        perf->value.it_value.tv_usec=PERFOMETER_MS*1000;
        break;
   case TIMER_NEWVALUE:
        if ( perf->paused_app ) {
	   tv_sec = PERFOMETER_SEC;
	   tv_usec= newsec*1000;
	   vtv_usec= newsec*1000;
	   vtv_sec= PERFOMETER_SEC;
	   return 0;
	}
	perf->value.it_interval.tv_sec = PERFOMETER_SEC;
        perf->value.it_interval.tv_usec =newsec*1000;
        perf->value.it_value.tv_sec = PERFOMETER_SEC;
        perf->value.it_value.tv_usec=newsec*1000;
        break;
   case TIMER_RESTART:
         if ( tv_sec == 0 && tv_usec == 0 && vtv_sec == 0 && vtv_usec == 0 ){
                perf->value.it_interval.tv_sec = PERFOMETER_SEC;
                perf->value.it_interval.tv_usec =PERFOMETER_MS*1000;
                perf->value.it_value.tv_sec = PERFOMETER_SEC;
                perf->value.it_value.tv_usec=PERFOMETER_MS*1000;
        }
        else {
                perf->value.it_interval.tv_sec = tv_sec;
                perf->value.it_interval.tv_usec = tv_usec;
                perf->value.it_value.tv_sec = vtv_sec;
                perf->value.it_value.tv_usec = vtv_usec;
        	tv_sec = 0;
        	tv_usec = 0;
        	vtv_sec = 0;
        	vtv_usec = 0;
        }
        break;
  }
#ifdef DEBUG
  fprintf(stdout,"Timer values: %d %d %d %d\n", 
	(int) perf->value.it_interval.tv_sec,
        (int) perf->value.it_interval.tv_usec, (int)perf->value.it_value.tv_sec,
        (int)perf->value.it_value.tv_usec);
#endif
  if ( setitimer( perfometer_timer, &(perf->value), NULL ) == -1 )
        perror("settimer()");
#else
  static UINT timeout=0;
  static int    started=0;
  static UINT_PTR mytimer=0;

  switch ( flag ) {
     case TIMER_PAUSE:
            timeout = perf->value;
            perf->value = 0;
            break;
    case TIMER_DEFAULT:
            perf->value = ((PERFOMETER_SEC*1000)+PERFOMETER_MS);
            break;
   case TIMER_NEWVALUE:
           if ( perf->paused_app ) {
	 	timeout = newsec;
		return 0;
	   } 
	   perf->value = ((PERFOMETER_SEC*1000)+newsec);
           break;
   case TIMER_RESTART:
           if ( timeout == 0 )
                   perf->value = ((PERFOMETER_SEC*1000)+PERFOMETER_MS);
           else
                   perf->value = timeout;
           break;
  }
  if ( perf->value && !started ) {
          mytimer=timeSetEvent( perf->value, 0, perfometer_handler, 0, TIME_PERIODIC );
          started = 1;
  }
  else if ( started && !perf->value ) {
          timeKillEvent( mytimer );
          started = 0;
  }
  else if ( started && perf->value ) {
          timeKillEvent( mytimer );
      mytimer=timeSetEvent( perf->value, 0, perfometer_handler, 0, TIME_PERIODIC );
          started = 1;
  }
#endif
  return 0;
}


/*
 * Parser for messages -KSL
 */

char * strip_word ( char *buf, char *buf_first, char delim ) {
  if ( !buf || buf[0] == '\0' ) {
        buf_first[0] = '\0';
        return NULL;
  }
  if ( *buf == delim )
        buf++;

  while ( *buf != '\0' && *buf != delim ) {
        *buf_first = *buf;
        buf_first++;
        buf++;
  }
  *buf_first = '\0';
  return buf;
}

/*
 * Write a message to the filedes, and setup the header to be sent
 */
int write_mesg(struct perfometer_data *perf, int filedes, int type, char *mesg ) {
    int     length,write_length;
    int     mysend;  /* Had to change this because microsoft doesn't like
		      * a send variable -KSL 
		      */
/*   static int is_connected=1; */
#ifndef WIN32
    ssize_t retval;
#else
    int retval;
#endif

    if ( !mesg )
        length = 0;
    else
        length = strlen(mesg);
    mysend = htonl( length );

#ifndef WIN32
    retval = write(filedes, (void*)&mysend,sizeof(int));
#else
    retval = send( filedes, (const char *) &mysend, sizeof(int), 0);
#endif

    if ( retval != sizeof(int) ) {
        fprintf(stderr, "Error sending on socket, sent %d bytes\n",retval);
#ifdef WIN32
	    exit(-1);
#endif
	_internal_stop_perfometer(perf);
        return(-1);
    }
    mysend = htonl(type);
#ifndef WIN32
    retval = write(filedes, (void*)&mysend,sizeof(int));
#else
    retval = send( filedes, (const char *) &mysend, sizeof(int), 0);
#endif
    if ( retval != sizeof(int) ) {
	fprintf(stderr, "Error sending on socket, 2nd int sent %d bytes\n",retval);
#ifdef WIN32
	    exit(-1);
#endif
    _internal_stop_perfometer(perf);
	return -1;
    }
    if ( mesg ) {
#ifndef WIN32
        retval = write(filedes,(void*)mesg,length);
#else
        retval = send(filedes, (const char *)mesg, length, 0 );
#endif
        if ( retval < 0 ) 
		goto socket_failure;
       if ( retval != length ) {
	  write_length = retval;
	  while ( write_length != length ) {
#ifndef WIN32
                retval = write(filedes,(void*)(mesg+write_length),length-write_length);
#else
                retval = send(filedes,(const char *)(mesg+write_length),length-write_length,0);
#endif
                if ( retval < 0 )
                        goto socket_failure;
                write_length+=retval;
	  }
       }
    }
#ifdef VERBOSE
    fprintf( stdout, "%d sending: %d|%d: %s\n", perf->my_uid,
                 length, type, mesg?mesg:" " );
#endif
    return 0;
socket_failure:
#ifdef WIN32
	    exit(-1);
#endif
    _internal_stop_perfometer(perf);
	return -1;
}

/*
 * Read in a message 
 */
int read_mesg ( struct perfometer_data *perf, int filedes, char *mesg ) {
    int type;
    int length=0,read_length;
    int readtmp;
    int i;
    char *tmp;
    char buf[128];
#ifndef WIN32
    ssize_t retval;
#else
    int retval;
#endif


#ifndef WIN32
    retval = read(filedes, (void*)&readtmp, sizeof(int));
#else
    retval = recv(filedes, (char *)&readtmp, sizeof(int), 0 );
#endif
    length = ntohl( readtmp );
    if ( retval != sizeof(int) ) {
        fprintf(stderr, "%d error recieving first int on socket, %d\n", perf->my_uid,retval);
#ifdef WIN32
	    exit(-1);
#endif
	_internal_stop_perfometer(perf);
	return -1;
    }

#ifndef WIN32
    retval = read(filedes, (void*)&readtmp, sizeof(int) );
#else
    retval = recv(filedes, (char *)&readtmp, sizeof(int), 0 );
#endif
    type = ntohl( readtmp );
    if ( retval != sizeof(int) ) {
        fprintf(stderr, "%d error recieving second int on socket, %d\n",perf->my_uid,retval);
#ifdef WIN32
	    exit(-1);
#endif	
	_internal_stop_perfometer(perf);
	return -1;
    }
    if ( length > 0 ) {
#ifndef WIN32
       retval = read(filedes, (void*)mesg,  length );
#else
       retval = recv(filedes, (char*)mesg, length, 0 );
#endif
       if ( retval < 0 ) {
          fprintf(stderr, "%d error recieving mesg on socket, %d\n",perf->my_uid,retval);
#ifdef WIN32
	    exit(-1);
#endif	
	  _internal_stop_perfometer(perf);
	  return -1;
       }
       if ( retval != length ){
	  read_length = retval;
	  while ( read_length != length ){
#ifndef WIN32
             retval=read(filedes, mesg+read_length,  length-read_length );
#else
             retval=recv(filedes, (mesg+read_length), length-read_length, 0 );
#endif
             if ( retval < 0 ) {
                 fprintf(stderr, "%d error recieving mesg on socket, %d\n",perf->my_uid,retval);
#ifdef WIN32
				 exit(-1);
#endif	
				 _internal_stop_perfometer(perf);
                 return -1;
             }
             read_length+=retval;
	  }
       }
       mesg[length] = '\0';
#ifdef VERBOSE
    if ( length > 0 )
       fprintf(stdout, "%d recieving message: %d|%d:%s\n", perf->my_uid, length, type, mesg );
    else
       fprintf( stdout, "%d recieving: %d|%d\n", perf->my_uid, length, type );
#endif
   }

   retval = type;
   switch ( type ) {
	case PERF_ACK:
           if ( length > 0 ) {
                tmp = mesg;
                tmp++;
                perf->my_uid = atoi(tmp);
           }
	   sprintf( mesg, "%d", type );
	   break;
	case PERF_PAUSE:
#ifdef DEBUG
           fprintf(stdout,"Pausing Application!\n");
#endif
	   perf->paused_app = 1;
	   change_timer( perf, TIMER_PAUSE, 0 );
#ifndef WIN32
#ifndef NO_ASYNCH
           sigpause(SIGIO);
#else
	   perfometer_check_messages_noasync( perf );
#endif
#else
           perfometer_check_messages_noasync( perf );
#endif
	   return retval;
	case PERF_STARTAPP:
#ifdef DEBUG
           fprintf(stdout,"Resuming Application!\n");
#endif
	   change_timer( perf, TIMER_RESTART, 0 );
	   perf->paused_app = 0;
	   break;
	case PERF_ABORT:
#ifdef DEBUG
           fprintf( stdout, "Application being shutdown by perfometer\n");
#endif
           exit( 1 );
        case PERF_CHANGE_EVENT:
#ifdef DEBUG
           printf("PERF_CHANGE_EVENT\n");
#endif
           perfometer_change_event( perf, mesg );
           break;
        case PERF_DATASTREAM:
           tmp=mesg;
           tmp++;
           for ( i = 0; i < length; i++,tmp++ ){
               if ( tmp[0]  == '^' ) {
                   tmp++;
                   break;
                }
           }
	   perf->data_stream = atoi( tmp );
#ifdef DEBUG
           fprintf( stdout, "DATASTREAM set to %d\n", perf->data_stream );
           fflush(stdout);
#endif
           break;
	case PERF_POLL:
	   tmp=mesg;
	   tmp++;
	   for ( i = 0; i < length; i++,tmp++ ){
               if ( tmp[0]  == '^' ) {
                   tmp++;
                   break;
                }
           }
	   change_timer( perf, TIMER_NEWVALUE, atoi(tmp) ); 
           sprintf(buf, "%s%d", PERF_DELIM, perf->my_uid);
           write_mesg(perf,perf->write_fd, PERF_OK, buf );
	break;
	default:
	   fprintf( stdout, "%d not implemented yet!\n", type );
	   break;
       }
    if ( type != PERF_STARTAPP && perf->paused_app ){
#ifdef DEBUG
	   fprintf(stdout,"Re-pausing application!!!\n");
#endif
#ifndef WIN32
#ifndef NO_ASYNCH
           sigpause(SIGIO);
#endif
#endif
    }
    return retval;
}

static void perfometer_check_messages_async(int sig, siginfo_t *info, void *ptr)
{
#ifdef THREADS
   struct perfometer_data *perf = lookup_thread_specific_data();
#else
   struct perfometer_data *perf = default_perfometer_data;
#endif
/*
  printf("%d,%d,%d,%d vs %d\n",info->si_signo,info->si_errno,info->si_code,info->si_fd,perf->read_fd);
*/
  read_mesg(perf, perf->read_fd, perf->buf );
}

#if (WIN32|NO_ASYNCH)
static void perfometer_check_messages_noasync(	struct perfometer_data *perf) {
        fd_set fdread;
    int ret;
        struct timeval timeout;

        timeout.tv_sec = 0;
	timeout.tv_usec = 1;

        FD_ZERO( &fdread );
        FD_SET ( perf->read_fd, &fdread );
#ifdef WIN32
        if ( !sig )
                ret = select(0, &fdread, NULL, NULL, NULL );
        else
                ret = select(0, &fdread, NULL, NULL, &timeout);
#else
        if ( !sig )
                ret = select((perf->read_fd+1), &fdread, NULL, NULL, NULL );
        else
                ret = select((perf->read_fd+1), &fdread, NULL, NULL, &timeout);
#endif
        if ( ret < 1 )
                return;
   read_mesg(perf, perf->read_fd, perf->buf );
}
#endif

int perfometer_change_event( struct perfometer_data *perf, char *mesg ) {
   char myerror[MAX_BUFFER_SIZE];
   char *tmp, *freethis;
   int check, preset,i;

   change_timer(perf, TIMER_PAUSE, 0 );
   if ( perf->EventSet != PAPI_NULL ) {
     PAPI_stop(perf->EventSet, perf->values );
     
     for ( i=0;i<perf->num_data_values;i++ )
     	PAPI_rem_event( &(perf->EventSet), perf->events[i] );


     if ( (check = PAPI_cleanup_eventset( &(perf->EventSet) )) < PAPI_OK ){
        PAPI_perror(check, perf->buf, 512 );
        fprintf(stderr,"PAPI_cleanup_eventset %s\n", perf->buf );
        sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
        write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
	return check;
     }
     PAPI_destroy_eventset( &(perf->EventSet) );
   }
   if ( (check = PAPI_create_eventset( &(perf->EventSet) )) != PAPI_OK ){
	PAPI_perror(check,perf->buf,512 );
	fprintf(stderr,"PAPI_create_eventset: %s\n", perf->buf );
        sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
        write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
	return check;
   }
  mesg =  strip_word( mesg, perf->buf, PERF_DELIM[0] ); /* This grabs ranks which we can ignore */
  mesg =  strip_word( mesg, perf->buf, PERF_DELIM[0] ); /* This grabs ranks which we can ignore */
  if ( !perf->buf || perf->buf[0] == '\0' ) {
     /* Error */
        fprintf(stderr,"Error in strip_word\n");
	return -1;
  }

  /* Parse the eventlist */
  tmp = strdup(perf->buf);
  perf->num_data_values = 0;
  freethis = tmp;
  while ( tmp ) {
        tmp = strip_word( tmp, perf->buf, '@' );
        if ( !perf->buf || perf->buf[0] == '\0' )
                break;
        check = PAPI_event_name_to_code( perf->buf, &preset );
        if ( check < PAPI_OK ) {
                PAPI_perror(check,perf->buf,512);
                fprintf(stderr,"PAPI_event_name_to_code: %s\n", perf->buf );
                sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
                write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
		free(freethis);
		return check;
        }
        check = PAPI_query_event(preset);
        if ( check < PAPI_OK ) {
                PAPI_perror(check,perf->buf,512);
                fprintf(stderr,"PAPI_query_event: %s\n", perf->buf );
                sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
                write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
		free(freethis);
		return check;
        }
        check = PAPI_add_event(&(perf->EventSet), preset );
        if ( check < PAPI_OK ) {
                PAPI_perror(check,perf->buf,512);
                fprintf(stderr,"PAPI_add_event: %s\n", perf->buf );
                sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
                write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
		free(freethis);
		return check;
        }
        perf->events[perf->num_data_values] = preset;
        perf->num_data_values++;
  }
  if ( (check = PAPI_start(perf->EventSet) ) < PAPI_OK ) {
	PAPI_perror(check,perf->buf,512);
	fprintf(stderr, "PAPI_start: %s\n", perf->buf );
        sprintf(myerror, "%s%d%s%s", PERF_DELIM, perf->my_uid,PERF_DELIM,perf->buf);
        write_mesg(perf,perf->write_fd, PERF_ERROR, myerror );
	free(freethis);
	return check;
  }
  sprintf(perf->buf,"^PAPI_start: %d", perf->my_uid );
  write_mesg(perf, perf->write_fd, PERF_OK, perf->buf );
  /* XXX */
  change_timer( perf, TIMER_RESTART, 0 ); 
#ifdef DEBUG
  fprintf(stdout,"CHANGED EVENT: num_events: %d\n", perf->num_data_values);
#endif
  free(freethis);
  return 0;
}

#ifndef WIN32
static void perfometer_handler( int sig ) {
#else
VOID CALLBACK perfometer_handler ( UINT wTimerID, UINT msg, DWORD dwuser, DWORD dw1, DWORD dw2) {
#endif
   int retval, i;
   char str[300], tmp[50];
   float total_real_time, total_virtual_time;
#ifdef SPEEDSTEP
   static unsigned long long num_cycles=0;
#endif
#ifdef NO_THREADS
   struct perfometer_data *perf = lookup_thread_specific_data();
#else
   struct perfometer_data *perf = default_perfometer_data;
#endif

   total_real_time=(float)((PAPI_get_real_usec() - perf->start_us)/1000000.0);
   total_virtual_time=(float)((PAPI_get_virt_usec() - perf->vstart_us)/1000000.0);

   retval = PAPI_read(perf->EventSet, perf->values); 
   PAPI_reset( perf->EventSet ); 

    if (retval < PAPI_OK){
        char errstring[1000];
        PAPI_perror(retval, errstring, 1000 );
        fprintf(stderr, "Error: %s\n", errstring );
	return;
    }

#ifdef SPEEDSTEP
    {
	unsigned long long temp_cycles;
	unsigned long long tmp;
        __asm__ __volatile__("rdtsc"
                            : "=A" (temp_cycles)
                            : /* no inputs */);
	tmp = llabs((temp_cycles-num_cycles));
	if ( tmp >  (temp_cycles*.2) ){
           float mhz;
	   mhz = perf->mhz;
	   perf->mhz = recalc_mhz();
#ifdef DEBUG
	   printf("Old mhz: %f New mhz: %f\n", mhz, perf->mhz );
#ifdef WIN32
	   printf("Temp_cycles: %I64d  Num_cycles: %I64d  Diff: %I64d\n",
		temp_cycles, num_cycles, tmp );
#else
	   printf("Temp_cycles: %lld  Num_cycles: %lld  Diff: %lld\n",
		temp_cycles, num_cycles, tmp );
#endif

#endif
	   num_cycles = temp_cycles;
	   if ( mhz != perf->mhz ) {
		char buf[200];
		sprintf(buf, "%s%d%s%f", PERF_DELIM,perf->my_uid,
			PERF_DELIM, perf->mhz );
		write_mesg(perf, perf->write_fd, PERF_MHZ, buf );
	   }
	}
    }
#endif
    if ( perf->data_stream ) {
	sprintf(str, "%s%d%s%d%s%d%s%d%s%s%s%.2f%s%.2f",              
		PERF_DELIM, perf->my_uid, PERF_DELIM, 1, PERF_DELIM, 
		perf->num_data_values, PERF_DELIM, perf->color, PERF_DELIM, 
		perf->label, PERF_DELIM, total_real_time,PERF_DELIM,
		total_virtual_time );
        for ( i=0; i<perf->num_data_values; i++ ) {
#ifdef WIN32
		sprintf(tmp, "%s%I64d", PERF_DELIM,perf->values[i] );
#else
		sprintf(tmp, "%s%lld", PERF_DELIM,perf->values[i] );
#endif
	  strcat( str, tmp );
         }
      write_mesg(perf, perf->write_fd, PERF_DATA, str );
    }
#ifdef DEBUG
    else
	fprintf(stdout,"Handler entered but data_stream set to 0.\n");
#endif
#if (WIN32|NO_ASYNCH)
    perfometer_check_messages(1);
#endif
    return;
}

#if 0
/* Removed -PJM */
/*
 * Force a time sample to be taken good for routines that
 * take less time than the current resolution. -KSL
 */
void internal_perfometer_update() {
  change_timer( perf, TIMER_PAUSE, 0 );
#ifndef WIN32
  perfometer_handler( 0 );
#else
  /* gotta implement the Windows call here... */
#endif
  change_timer( perf, TIMER_RESTART, 0 );
  return;
}
#endif
/*
 * Change the color scheme and label
 */
void _internal_mark_perfometer( void *perfometer, int color, char *label ){
  struct perfometer_data *perf = (struct perfometer_data *)perfometer;
   if (perf == NULL)
   return;
   perf->color = color;
   if ( label && label[0] != '\0' ){
        if ( strlen(label) < 128 )
                strcpy( perf->label, label );
        else {
                strncpy( perf->label, label, 128 );
                perf->label[127] = '\0';
        }
   }
   else
        strcpy( perf->label, "Unlabeled");
   return;
}

/*
 * Send a message about the program ending
 */
void _internal_stop_perfometer(void *perfometer ) {
  struct perfometer_data *perf = (struct perfometer_data *)perfometer;
    if (perf == NULL)
    return;
    sprintf(perf->buf, "%d", perf->my_uid );
    write_mesg(perf, perf->write_fd, PERF_END, perf->buf );
    PAPI_stop( perf->EventSet, perf->values );
    PAPI_cleanup_eventset(&(perf->EventSet) );
    PAPI_destroy_eventset(&(perf->EventSet) );
    change_timer( perf,TIMER_PAUSE, 0 );
#ifndef WIN32
    signal( perfometer_signal, SIG_IGN);
#ifndef NO_ASYNCH
    signal( perfometer_readsig, SIG_IGN);
#endif
    close ( perf->write_fd );
    close( perf->read_fd );
#else
        closesocket( perf->write_fd );
        closesocket( perf->read_fd );
#endif
    free( perf->user_name );
    free( perf->my_hostip);
    return;
}

#ifdef SPEEDSTEP
float recalc_mhz() {
   char *s;
   char maxargs[PAPI_MAX_STR_LEN];
   float mhz, dec_mhz;
   FILE *f;
   int tmp;

  if ((f = fopen("/proc/cpuinfo", "r")) == NULL)
    return -1;
  s = (char *) search_cpu_info(f,"cpu MHz",maxargs);
  if (s)
    sscanf(s+1, "%f", &dec_mhz);
  fclose(f);
  return dec_mhz;
  mhz = calc_mhz();
  if ( dec_mhz < mhz )
	tmp = mhz;
  else
	tmp = dec_mhz;
  return tmp;
}

static float calc_mhz(void)
{
  unsigned long long ostamp;
  unsigned long long stamp;
  float correction = 4000.0, mhz;

  /* Warm the cache */

  ostamp = get_cycles();
  usleep(1);
  stamp = get_cycles();
  stamp = stamp - ostamp;
  mhz = (float)stamp/(float)(1000000.0 + correction);

  ostamp = get_cycles();
  sleep(1);
  stamp = get_cycles();
  stamp = stamp - ostamp;
  mhz = (float)stamp/(float)(1000000.0 + correction);

  return(mhz);
} 


static inline unsigned long long get_cycles (void)
{
        unsigned long long ret;
        __asm__ __volatile__("rdtsc"
                            : "=A" (ret)
                            : /* no inputs */);
        return ret;
}


inline static char *search_cpu_info(FILE *f, char *search_str, char *line)
{
  /* This code courtesy of our friends in Germany. Thanks Rudolph Berrendorf! */  /* See the PCL home page for the German version of PAPI. */

  char *s;

  while (fgets(line, 256, f) != NULL)
    {
      if (strstr(line, search_str) != NULL)
        {
          /* ignore all characters in line up to : */
          for (s = line; *s && (*s != ':'); ++s)
            ;
          if (*s)
            return(s);
        }
    }
  return(NULL);

  /* End stolen code */
}
#endif

